@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2008 Gustavo Noronha Silva
+ * Copyright (C) 2008, 2010 Gustavo Noronha Silva
  *
  * This file is part of the Gksu PolicyKit library.
  *
@@ -29,13 +29,13 @@
 #include <dbus/dbus-glib.h>
 #include <dbus/dbus-glib-lowlevel.h>
 #include <dbus/dbus.h>
-#include <polkit-dbus/polkit-dbus.h>
 
 #define SN_API_NOT_YET_FROZEN
 #include <libsn/sn.h>
 #include <gdk/gdk.h>
 #include <gdk/gdkx.h>
 
+#include <gksu-process-error.h>
 #include <gksu-environment.h>
 #include <gksu-write-queue.h>
 #include <gksu-marshal.h>
@@ -431,7 +431,7 @@ gksu_process_close_server_fd(GksuProcess *self, guint fd)
 {
   GksuProcessPrivate *priv = GKSU_PROCESS_GET_PRIVATE(self);
   GError *error = NULL;
-  
+
   dbus_g_proxy_call(priv->server, "CloseFD", &error,
                     G_TYPE_UINT, priv->cookie,
                     G_TYPE_INT, fd,
@@ -540,7 +540,7 @@ gksu_process_stdin_ready_to_send_cb(GIOChannel *channel, GIOCondition condition,
  * @arguments array must have the command to be executed at its first
  * position, followed by the command's arguments; it must also contain
  * a %NULL at its last position.
- * 
+ *
  * Returns: a new instance of #GksuProcess
  */
 GksuProcess*
@@ -637,19 +637,7 @@ gksu_process_spawn_async_with_pipes(GksuProcess *self, gint *standard_input,
 
   if(internal_error)
     {
-      if(g_str_has_prefix(internal_error->message, "auth_"))
-	{
-	  DBusError dbus_error;
-	  dbus_error_init(&dbus_error);
-	  if (polkit_auth_obtain("org.gnome.gksu.spawn",
-                                 0, getpid(), &dbus_error))
-            return gksu_process_spawn_async_with_pipes(self, standard_input,
-                                                       standard_output, standard_error,
-                                                       error);
-	}
-      else
-        g_propagate_error(error, internal_error);
-
+      g_propagate_error(error, internal_error);
       return FALSE;
     }
 
@@ -664,12 +652,12 @@ gksu_process_spawn_async_with_pipes(GksuProcess *self, gint *standard_input,
                                 standard_input,
                                 TRUE);
 
-      priv->stdin_source_id = 
+      priv->stdin_source_id =
         g_io_add_watch(priv->stdin_channel, G_IO_IN|G_IO_PRI,
                        (GIOFunc)gksu_process_stdin_ready_to_send_cb,
                        (gpointer)self);
 
-      priv->stdin_mirror_id = 
+      priv->stdin_mirror_id =
         g_io_add_watch(priv->stdin_mirror, G_IO_HUP|G_IO_NVAL,
                        (GIOFunc)gksu_process_stdin_mirror_hangup_cb,
                        (gpointer)self);
@@ -749,7 +737,7 @@ sync_handle_exited(GksuProcess *self, gint status, SyncRunInfo *sri)
 }
 
 /**
- * gksu_process_spawn_async
+ * gksu_process_spawn_sync
  * @self: a #GksuProcess instance
  * @status: return location for the child's exit status, as returned
  * by waitpid(2)
@@ -796,3 +784,37 @@ gksu_process_spawn_sync(GksuProcess *self, gint *status, GError **error)
 
   return retval;
 }
+
+/**
+ * gksu_process_send_signal
+ * @self: a #GksuProcess instance
+ * @signum: a signal number to send to the process
+ * @error: return location for a #GError
+ *
+ * Asks the Gksu controlling server to send a signal to the process
+ * represented by the #GksuProcess.
+ *
+ * Since: 0.0.2
+ *
+ * Returns: %FALSE if @error is set, %TRUE if all went well
+ */
+gboolean
+gksu_process_send_signal(GksuProcess *self, gint signum, GError **error)
+{
+  GksuProcessPrivate *priv = GKSU_PROCESS_GET_PRIVATE(self);
+  GError *internal_error = NULL;
+
+  dbus_g_proxy_call(priv->server, "SendSignal", &internal_error,
+                    G_TYPE_UINT, priv->cookie,
+                    G_TYPE_INT, signum,
+                    G_TYPE_INVALID,
+                    G_TYPE_INVALID);
+
+  if(internal_error)
+    {
+      g_propagate_error(error, internal_error);
+      return FALSE;
+    }
+
+  return TRUE;
+}
